unit can.gw;

interface

uses
  ctypes, can;

(* SPDX-License-Identifier: ((GPL-2.0-only WITH Linux-syscall-note) OR BSD-3-Clause) *)
(*
 * linux/can/gw.h
 *
 * Definitions for CAN frame Gateway/Router/Bridge
 *
 * Author: Oliver Hartkopp <oliver.hartkopp@volkswagen.de>
 * Copyright (c) 2011 Volkswagen Group Electronic Research
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Volkswagen nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ('GPL') version 2, in which case the provisions of the
 * GPL apply INSTEAD OF those given above.
 *
 * The provided data structures and external interfaces from this code
 * are not restricted to be used by modules with a GPL compatible license.
 *
 *  SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING  ANY WAY  OF THE USE
 * OF  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *)

type
  rtcanmsg = record
    can_family: cuint8;
    gwtype:     cuint8;
    flags:      cuint16;
  end;

const
  (* CAN gateway types *)
  CGW_TYPE_UNSPEC  = 0;
  CGW_TYPE_CAN_CAN = 1;	(* CAN->CAN routing *)
  __CGW_TYPE_MAX   = 2;

  CGW_TYPE_MAX = __CGW_TYPE_MAX - 1;

  (* CAN rtnetlink attribute definitions *)
  CGW_UNSPEC    =  0;
  CGW_MOD_AND   =  1;  (* CAN frame modification binary AND *)
  CGW_MOD_OR    =  2;  (* CAN frame modification binary OR *)
  CGW_MOD_XOR   =  3;  (* CAN frame modification binary XOR *)
  CGW_MOD_SET   =  4;  (* CAN frame modification set alternate values *)
  CGW_CS_XOR    =  5;  (* set data[] XOR checksum into data[index] *)
  CGW_CS_CRC8   =  6;  (* set data[] CRC8 checksum into data[index] *)
  CGW_HANDLED   =  7;  (* number of handled CAN frames *)
  CGW_DROPPED   =  8;  (* number of dropped CAN frames *)
  CGW_SRC_IF    =  9;  (* ifindex of source network interface *)
  CGW_DST_IF    = 10;  (* ifindex of destination network interface *)
  CGW_FILTER    = 11;  (* specify struct can_filter on source CAN device *)
  CGW_DELETED   = 12;  (* number of deleted CAN frames (see max_hops param) *)
  CGW_LIM_HOPS  = 13;  (* limit the number of hops of this specific rule *)
  CGW_MOD_UID   = 14;  (* user defined identifier for modification updates *)
  CGW_FDMOD_AND = 15;  (* CAN FD frame modification binary AND *)
  CGW_FDMOD_OR  = 16;  (* CAN FD frame modification binary OR *)
  CGW_FDMOD_XOR = 17;  (* CAN FD frame modification binary XOR *)
  CGW_FDMOD_SET = 18;  (* CAN FD frame modification set alternate values *)
  __CGW_MAX     = 19;

  CGW_MAX = __CGW_MAX - 1;

  CGW_FLAGS_CAN_ECHO       = $01;
  CGW_FLAGS_CAN_SRC_TSTAMP = $02;
  CGW_FLAGS_CAN_IIF_TX_OK  = $04;
  CGW_FLAGS_CAN_FD         = $08;

  CGW_MOD_FUNCS = 4;	(* AND OR XOR SET *)

  (* CAN frame elements that are affected by curr. 3 CAN frame modifications *)
  CGW_MOD_ID    = $01;
  CGW_MOD_DLC   = $02;          (* contains the data length in bytes *)
  CGW_MOD_LEN   = CGW_MOD_DLC;  (* CAN FD length representation      *)
  CGW_MOD_DATA  = $04;
  CGW_MOD_FLAGS = $08;          (* CAN FD flags                      *)

  CGW_FRAME_MODS = 4;           (* ID DLC/LEN DATA FLAGS             *)

  MAX_MODFUNCTIONS = CGW_MOD_FUNCS * CGW_FRAME_MODS;

type
  cgw_frame_mod = packed record
    cf:      can_frame;
    modtype: cuint8;
  end;

  cgw_fdframe_mod = packed record
    cf:      canfd_frame;
    modtype: cuint8;
  end;

const
  CGW_MODATTR_LEN   = SizeOf(cgw_frame_mod);
  CGW_FDMODATTR_LEN = SizeOf(cgw_fdframe_mod);

type
  cgw_csum_xor = packed record
    from_idx:     cint8;
    to_idx:       cint8;
    result_idx:   cint8;
    init_xor_val: cuint8;
  end;

  cgw_csum_crc8 = packed record
    from_idx:      cint8;
    to_idx:        cint8;
    result_idx:    cint8;
    init_crc_val:  cuint8;
    final_xor_val: cuint8;
    crctab:        array[0..255] of cuint8;
    profile:       cuint8;
    profile_data:  array[0..19] of cuint8;
  end;

const
  (* length of checksum operation parameters. idx = index in CAN frame data[] *)
  CGW_CS_XOR_LEN  = SizeOf(cgw_csum_xor);
  CGW_CS_CRC8_LEN = SizeOf(cgw_csum_crc8);

  (* CRC8 profiles (compute CRC for additional data elements - see below) *)
  CGW_CRC8PRF_UNSPEC    = 0;
  CGW_CRC8PRF_1U8       = 1;  (* compute one additional u8 value         *)
  CGW_CRC8PRF_16U8      = 2;  (* u8 value table indexed by data[1] & 0xF *)
  CGW_CRC8PRF_SFFID_XOR = 3;  (* (can_id & 0xFF) ^ (can_id >> 8 & 0xFF)  *)
  __CGW_CRC8PRF_MAX     = 4;

  CGW_CRC8PRF_MAX = __CGW_CRC8PRF_MAX - 1;

  (*
   * CAN rtnetlink attribute contents in detail
   *
   * CGW_XXX_IF (length 4 bytes):
   * Sets an interface index for source/destination network interfaces.
   * For the CAN^.CAN gwtype the indices of _two_ CAN interfaces are mandatory.
   *
   * CGW_FILTER (length 8 bytes):
   * Sets a CAN receive filter for the gateway job specified by the
   *  can_filter described in include/linux/can.h
   *
   * CGW_MOD_(AND or OR or XOR or SET) (length 17 bytes):
   * Specifies a modification that's done to a received CAN frame before it is
   * send out to the destination interface.
   *
   * < can_frame> data used as operator
   * <u8> affected CAN frame elements
   *
   * CGW_LIM_HOPS (length 1 byte):
   * Limit the number of hops of this specific rule. Usually the received CAN
   * frame can be processed as much as 'max_hops' times (which is given at module
   * load time of the can-gw module). This value is used to reduce the number of
   * possible hops for this gateway rule to a value smaller then max_hops.
   *
   * CGW_MOD_UID (length 4 bytes):
   * Optional non-zero user defined routing job identifier to alter existing
   * modification settings at runtime.
   *
   * CGW_CS_XOR (length 4 bytes):
   * Set a simple XOR checksum starting with an initial value into
   * data[aResult-idx] using data[start-idx] .. data[end-idx]
   *
   * The XOR checksum is calculated like this:
   *
   * xor = init_xor_val
   *
   * for (i = from_idx .. to_idx)
   *      xor:= mod xor can_frame.data[i]
   *
   * can_frame.data[ result_idx ] = xor
   *
   * CGW_CS_CRC8 (length 282 bytes):
   * Set a CRC8 value into data[aResult-idx] using a given 256 byte CRC8 table,
   * a given initial value and a defined input data[start-idx] .. data[end-idx].
   * Finally the aResult value is XOR'ed with the final_xor_val.
   *
   * The CRC8 checksum is calculated like this:
   *
   * crc = init_crc_val
   *
   * for (i = from_idx .. to_idx)
   *      crc = crctab[ crc xor can_frame.data[i] ]
   *
   * can_frame.data[ result_idx ] = crc xor final_xor_val
   *
   * The calculated CRC may contain additional source data elements that can be
   * defined in the handling of 'checksum profiles' e.g. shown in AUTOSAR specs
   * like http://www.autosar.org/download/R4.0/AUTOSAR_SWS_E2ELibrary.pdf
   * E.g. the profile_data[] may contain additional u8 values (called DATA_IDs)
   * that are used depending on counter values inside the CAN frame data[].
   * So  only three profiles have been implemented for illustration.
   *
   * Remark: In general the attribute data is a linear buffer.
   *         Beware of sending unpacked or aligned structs not
   *)

implementation

end.

