unit can.hl;

{$mode objfpc}
{$modeswitch typehelpers}
{$modeswitch advancedrecords}
{$inline on}

interface

uses
  ctypes, baseunix, sockets, can, can.bcm;

// +=====+=====+=====+============================================+
// | 31  | 30  | 29  |                   28..0                    |
// |-----+-----+-----+--------------------------------------------|
// | EFF | RTR | ERR |                   CAN ID                   |   <= CAN frame
// +=====+=====+=====+============================================+
// | 31  | 30  | 29  | 28..26   |           25..8          | 7..0 |
// |-----+-----+-----+----------+--------------------------+------|
// | EFF | RTR | ERR | PRIORITY |            PGN           |  SA  |   <= J1939 frame
// +=====+=====+=====+==========+==========================+======+
// | 31  | 30  | 29  | 28..26   | 25 | 24 | 23..16 | 15..8 | 7..0 |
// |-----+-----+-----+----------+----+----+--------+-------+------|
// | EFF | RTR | ERR | PRIORITY | R  | DP |   PF   |  PS   |  SA  |   <= J1939 frame with PGN parts
// +=====+=====+=====+==========+====+====+========+=======+======+

const
  S_IRUSR =  1 shl 3;

type
  // Controller Area Network Identifier structure
  TCanID = bitpacked record case byte of // high level for canid_t
    1: (Value: 0..(1 shl 29)-1; // bit 0-28 : CAN identifier (11/29 bit)
        ERR:   boolean;         // bit 29   : error message frame flag (0 = data frame, 1 = error message)
        RTR:   boolean;         // bit 30   : remote transmission request flag (1 = rtr frame)
        EFF:   boolean;);       // bit 31   : extended frame format flag (0 = standard 11 bit, 1 = extended 29 bit)
    2: (Bit:   bitpacked array[0..31] of boolean); // bit access helper
    3: (Full:  cuint32);                           // full access helper
  end;

  // CAN ID based filter in can_register()
  TCanFilter = record // high level for can_filter
    ID:   TCanID;     // relevant bits of CAN ID which are not masked out
    Mask: canid_t;    // CAN mask
                      // A filter matches, when <received_can_id> and Mask = ID and Mask
                      // The filter can be inverted (CAN_INV_FILTER bit set in ID) or it can
                      // filter for error message frames (CAN_ERR_FLAG bit set in mask).
  end;

  // Controller Area Network Error Message Frame Mask structure
  // bit 0-28	: error class mask (see include/uapi/linux/can/error.h)
  // bit 29-31	: set to zero
  // error class (mask) in can_id
  TCanErrMask = bitpacked record case byte of // high level for can_err_mask_t
    1: (ErrTxTimeout: boolean; // CAN_ERR_TX_TIMEOUT = $00000001; // TX timeout (by netdevice driver)
        ErrLostArb:   boolean; // CAN_ERR_LOSTARB    = $00000002; // lost arbitration    / data[0]
        ErrCtrl:      boolean; // CAN_ERR_CRTL       = $00000004; // controller problems / data[1]
        ErrProt:      boolean; // CAN_ERR_PROT       = $00000008; // protocol violations / data[2..3]
        ErrTrx:       boolean; // CAN_ERR_TRX        = $00000010; // transceiver status  / data[4]
        ErrAck:       boolean; // CAN_ERR_ACK        = $00000020; // received no ACK on transmission
        ErrBusOff:    boolean; // CAN_ERR_BUSOFF     = $00000040; // bus off
        ErrBusError:  boolean; // CAN_ERR_BUSERROR   = $00000080; // bus error (may flood!)
        ErrRestarted: boolean; // CAN_ERR_RESTARTED  = $00000100; // controller restarted
        __Dummy:      0..(1 shl 20)-1;                            // 20 bits not defined yet
        Zeros:        0..7;);                                     // set to zero
    2: (Bit:          bitpacked array[0..31] of boolean);         // bit access helper
    3: (Value:        cuint32);                                   // full access helper
  end;

  {$PACKRECORDS 8}
    // Basic CAN frame structure
    TCanFrame = record // high level for can_frame
      ID:     TCanID;  // 32 bit CAN_ID + EFF/RTR/ERR flags
      DLC:    cuint8;  // frame payload length in byte (0 .. CAN_MAX_DLEN)
      __pad:  cuint8;  // padding
      __res0: cuint8;  // reserved / padding
      __res1: cuint8;  // reserved / padding
      case byte of
        1: (Data:      array[0..CAN_MAX_DLEN-1]         of cuint8);  // originally defined in can.h
        2: (DataText:  array[0..CAN_MAX_DLEN-1]         of char);    // added helper for string assignment
        3: (DataByte:  array[0..CAN_MAX_DLEN-1]         of byte);    // added helper for byte assignment (same as cuint8)
        4: (DataWord:  array[0..(CAN_MAX_DLEN div 2)-1] of word);    // added helper for word assignment
        5: (DataDWord: array[0..(CAN_MAX_DLEN div 4)-1] of dword);   // added helper for double word assignment
        6: (DataQWord: array[0..(CAN_MAX_DLEN div 8)-1] of qword);   // added helper for quad word assignment
    end;
  {$PACKRECORDS DEFAULT}

  {$PACKRECORDS 8}
    // CAN flexible data rate frame structure
    TCanFDFrame = record // high level for canfd_frame
      ID: TCanID;        // 32 bit CAN_ID + EFF/RTR/ERR flags
      Len:    cuint8;    // frame payload length in byte (0 .. CANFD_MAX_DLEN)
      Flags:  cuint8;    // additional flags for CAN FD
      __res0: cuint8;    // reserved / padding
      __res1: cuint8;    // reserved / padding
      case byte of
        1: (Data:      array[0..CANFD_MAX_DLEN-1]         of cuint8);  // originally defined in can.h
        2: (DataText:  array[0..CANFD_MAX_DLEN-1]         of char);    // added helper for string assignment
        3: (DataByte:  array[0..CANFD_MAX_DLEN-1]         of byte);    // added helper for byte assignment (same as cuint8)
        4: (DataWord:  array[0..(CANFD_MAX_DLEN div 2)-1] of word);    // added helper for word assignment
        5: (DataDword: array[0..(CANFD_MAX_DLEN div 4)-1] of dword);   // added helper for double word assignment
        6: (DataQword: array[0..(CANFD_MAX_DLEN div 8)-1] of qword);   // added helper for quad word assignment
    end;
  {$PACKRECORDS DEFAULT}

  // SAE J1939 message format: https://www.kernel.org/doc/html/latest/networking/j1939.html
  TJ1939PS = record case byte of // J1939 PS (PDU specific), as used in J1939 PGN
    1: (Value: cuint8);   // full access helper
    2: (DA:    cuint8);   // DA (Destination Address), PDU1 Format (specific) (peer to peer), used when PF < 240 (0..$EF)
    3: (GE:    cuint8);   // GE (Group Extenstion),    PDU2 Format (global) (broadcast), used when PF >= 240 ($F0..$FF)
  end;

  TJ1939PGN = bitpacked record case byte of // J1939 PGN (Parameter Group Number), as used in J1939 29 bit CAN ID
    1: (Value: 0..(1 shl 18)-1;); // PGN value access helper
    2: (PS:    TJ1939PS;          // 8 bit: PS in PDU2 , else 0 (PDU specific)
        PF:    cuint8;            // 8 bit: PF (PDU format). PDU1 (0..239) indicates a destination address in PS, PDU2 (250-255) indicates extension to PDU Format (PF).
        DP:    boolean;           // 1 bit: DP (Data Page), page selector for PDU (Protocol Data Unit) Format (PF) field. Currently at 0, pointing to Page 0. Page 1 is for future purposes)
        R:     boolean);          // 1 bit: R  (Reserved), should always be set to 0 when transmitting messages
    3: (Full: cuint32;);          // full access helper (because sockaddr_can.pgn in can.h is cuint32)
  end;

  TJ1939SimplePGN = 0..(1 shl 18)-1; // when we can not use TJ1939PGN (bits are lost when used in records because of byte alignment)

  TJ1939CanID = bitpacked record case byte of // J1939 29 bit CAN ID
    1: (SA:       cuint8;            // SA (Source Address)
        case byte of
          11: (PGN:      TJ1939SimplePGN; // PGN 18 bits
               Priority: 0..7;);     // Priority, highest is  0 (%000), lowest is 7 (%111);
          12: (PGN_PS:   TJ1939PS;   // 8 bit: PS in PDU2 , else 0 (PDU specific)
               PGN_PF:   cuint8;     // 8 bit: PF (PDU format). PDU1 (0..239) indicates a destination address in PS, PDU2 (250-255) indicates extension to PDU Format (PF)
               PGN_DP:   boolean;    // 1 bit: DP (Data Page), page selector for PDU (Protocol Data Unit) Format (PF) field. Currently at 0, pointing to Page 0. Page 1 is for future purposes.
               PGN_R:    boolean;)); // 1 bit: R  (Reserved), should always be set to 0 when transmitting messages
               // Priority: 0..7;);  // Priority bits are already defined above
        // ERR, RTR and EFF bits used here are the same as those same bits below (declaring them twice is not allowed)
    2: (Value:    0..(1 shl 29)-1;   // bit 0-28 : CAN identifier (11/29 bit)
        ERR:      boolean;           // bit 29   : error message frame flag (0 = data frame, 1 = error message)
        RTR:      boolean;           // bit 30   : remote transmission request flag (1 = rtr frame)
        EFF:      boolean;);         // bit 31   : extended frame format flag (0 = standard 11 bit, 1 = extended 29 bit)
    3: (Bit:      bitpacked array[0..31] of boolean); // bit access helper
    4: (Full:     cuint32);                           // full access helper
  end;

  TJ1939Addr = cuint8; // 1 byte address

  TJ1939Name = bitpacked record case byte of
    1: (IdentityNumber:          0..(1 shl 21)-1; // bit 00..20
        ManufacturerCode:        0..(1 shl 11)-1; // bit 21..31
        ECUInstance:             0..(1 shl  3)-1; // bit 32..34
        aFunctionInstance:       0..(1 shl  5)-1; // bit 35..39
        aFunction:               0..(1 shl  8)-1; // bit 40..47
        Reserved:                boolean;         // bit 48
        VehicleSystem:           0..(1 shl  7)-1; // bit 49-55
        VehicleSystemInstance:   0..(1 shl  4)-1; // bit 56-59
        IndustryGroup:           0..(1 shl  3)-1; // bit 60-62
        ArbitraryAddressCapable: boolean;);       // bit 63
    2: (Value: cuint64);
    3: (Text:  array[0..7] of char);  // added helper for string assignment
    4: (Byte:  array[0..7] of byte);  // added helper for byte assignment (same as cuint8)
    5: (Word:  array[0..3] of word);  // added helper for word assignment
    6: (DWord: array[0..1] of dword); // added helper for double word assignment
    7: (QWord: array[0..0] of qword); // added helper for quad word assignment
  end;

  // sockaddr structure for CAN sockets
  TCanSockAddr = record               // high level for sockaddr_can
    CanFamily:  __kernel_sa_family_t; // address family number AF_CAN
    CanIfIndex: cint;                 // CAN network interface index (Original type was integer. Is cint (which equals cint32) a proper translation?)
    CanAddr: record case byte of      // protocol specific address information
      1: (TP: record
            RxID: canid_t;
            TxID: canid_t;
          end;);
      2: (J1939: record
            Name: TJ1939Name; // 8 byte name when using dynamic addressing
            PGN:  TJ1939PGN;  // Parameter Group Number
            Addr: TJ1939Addr; // 1 byte address
          end;);
      // reserved for future CAN protocols address information
    end;
    procedure Clear;
  end;

  TCanIfreq = record // high level for ifreq from can.if.inc
    case byte of
      1: (  // allow indirect data usage through ifr_ifrn and ifr_ifru structures
            IFRN : record
              Name : array[0..(IFNAMSIZ)-1] of char;
            end;
            IFRU : record
              case byte of
                1 : ( Addr      : sockaddr );
                2 : ( DSTAddr   : sockaddr );
                3 : ( BroadAddr : sockaddr );
                4 : ( NetMask   : sockaddr );
                5 : ( HwAddr    : sockaddr );
                6 : ( Flags     : cshort );
                7 : ( IValue    : cint );
                8 : ( MTU       : cint );
                9 : ( Map       : ifmap );
                10: ( Slave     : array[0..(IFNAMSIZ)-1] of char );
                11: ( NewName   : array[0..(IFNAMSIZ)-1] of char );
                12: ( Data      : __caddr_t );
                13: ( Settings  : if_settings );
            end;
         );
      2: (  // allow direct data usage without ifr_ifrn and ifr_ifru structures (slightly different field naming)
            Name : array[0..(IFNAMSIZ)-1] of char;                 (* interface name        *)
            case byte of
              1 : ( Addr      : sockaddr );                        (* address               *)
              2 : ( DSTAddr   : sockaddr );                        (* other end of p-p lnk  *)
              3 : ( BroadAddr : sockaddr );                        (* broadcast address     *)
              4 : ( NetMask   : sockaddr );                        (* interface net mask    *)
              5 : ( HwAddr    : sockaddr );                        (* MAC address           *)
              6 : ( Flags     : cshort );                          (* flags                 *)
              7 : ( Metric    : cint );                            (* metric                *)
              8 : ( MTU       : cint );                            (* mtu                   *)
              9 : ( Map       : ifmap );                           (* device map            *)
              10: ( Slave     : array[0..(IFNAMSIZ)-1] of char );  (* slave device          *)
              11: ( NewName   : array[0..(IFNAMSIZ)-1] of char );  (* New name              *)
              12: ( Data      : __caddr_t );                       (* for use by interface  *)
              13: ( Bandwidth : cint );                            (* link bandwidth        *)
              14: ( IfIndex   : cint );                            (* interface index       *)
              15: ( QLen      : cint );                            (* Queue length          *)
              16: ( Settings  : if_settings );                     (* Device/proto settings *)
         );
  end;

  // helpers for can.bcm

  TBcmTimeVal = record                    // high level for bcm_timeval in can.bcm
    Sec:  clong;
    uSec: clong;
  end;

  TBcmMsgHead = record                    // high level for bcm_msg_head in can.bcm
    OpCode:  cuint32;
    Flags:   cuint32;
    Count:   cuint32;
    IVal1:   TBcmTimeVal;
    IVal2:   TBcmTimeVal;
    ID:      TCanID;
    NFrames: cuint32;
    Frames:  array[0..0] of TCanFrame;    // frames[] removed because for BCM cyclic demo when more then 1 frame is needed we already
  end;                                    // had frames[0] in original bcm_msg_head therefore we had to populate 2 frame structures

  TBcmMsgHeadWithoutFrames = record       // high level for bcm_msg_head_without_frames in can.bcm
    OpCode:  cuint32;
    Flags:   cuint32;
    Count:   cuint32;
    IVal1:   TBcmTimeVal;
    IVal2:   TBcmTimeVal;
    ID:      TCanID;
    NFrames: cuint32;
    // frames:  array[0..0] of can_frame; // frames[] removed because for BCM cyclic demo when more then 1 frame is needed we already
  end;                                    // had frames[0] in original bcm_msg_head therefore we had to populate 2 frame structures

  // Protocol families helper, as defined in socketsh.inc
  TProtocolFamily = ( pfUnspec     = PF_UNSPEC,
                      pfUnix       = PF_UNIX,
                      pfLocal      = PF_LOCAL,
                      pfInet       = PF_INET,
                      pfAX25       = PF_AX25,
                      pfIPX        = PF_IPX,
                      pfAppleTalk  = PF_APPLETALK,
                      pfNetRom     = PF_NETROM,
                      pfBridge     = PF_BRIDGE,
                      pfATMPVC     = PF_ATMPVC,
                      pfX25        = PF_X25,
                      pfInet6      = PF_INET6,
                      pfRose       = PF_ROSE,
                      pfDecNet     = PF_DECnet,
                      pfNetBEUI    = PF_NETBEUI,
                      pfSecurity   = PF_SECURITY,
                      pfKey        = PF_KEY,
                      pfNetLink    = PF_NETLINK,
                      pfRoute      = PF_ROUTE,
                      pfPacket     = PF_PACKET,
                      pfASH        = PF_ASH,
                      pfEcoNet     = PF_ECONET,
                      pfATMSVC     = PF_ATMSVC,
                      pfRDS        = PF_RDS,
                      pfSNA        = PF_SNA,
                      pfIRDA       = PF_IRDA,
                      pfPPOX       = PF_PPPOX,
                      pfWanPipe    = PF_WANPIPE,
                      pfLLC        = PF_LLC,
                      pfIB         = PF_IB,
                      pfMPLS       = PF_MPLS,
                      pfCAN        = PF_CAN,
                      pfTIPC       = PF_TIPC,
                      pfBluetooth  = PF_BLUETOOTH,
                      pfIUCV       = PF_IUCV,
                      pfRXRPC      = PF_RXRPC,
                      pfISDN       = PF_ISDN,
                      pfPHONET     = PF_PHONET,
                      pfIEEE802154 = PF_IEEE802154,
                      pfCAIF       = PF_CAIF,
                      pfALG        = PF_ALG,
                      pfNFC        = PF_NFC,
                      pfVSOCK      = PF_VSOCK,
                      pfKCM        = PF_KCM,
                      pfQIPCRTR    = PF_QIPCRTR,
                      pfSMC        = PF_SMC,
                      pfXDP        = PF_XDP,
                      pfMAX        = PF_MAX );

  TCanHandle = cint;

  // helpers for socketsh.inc
  function  CanSocket      (domain: cint; xtype: cint; protocol: cint): TCanHandle; overload; inline;
  function  CanSocket      (domain: TProtocolFamily; xtype: cint; protocol: cint): TCanHandle; overload; inline;
  function  CanSocket      (xtype: cint; protocol: cint): TCanHandle; overload; inline;
  function  CanRecv        (Handle: TCanHandle; buf: pointer; len: size_t; flags: cint): ssize_t; inline;
  function  CanRecvFrom    (Handle: TCanHandle; buf: pointer; len: size_t; flags: cint; from: psockaddr; fromlen : psocklen): ssize_t;
  function  CanSendTo      (Handle: TCanHandle; msg: pointer; len: size_t; flags: cint; tox: psockaddr; tolen: tsocklen): ssize_t; inline;
  function  CanBind        (Handle: TCanHandle; addrx: psockaddr; addrlen: tsocklen): cint; overload; inline;
  function  CanBind        (Handle: TCanHandle; CanSockAddr: TCanSockAddr; addrlen: tsocklen): cint; overload; inline;
  function  CanBind        (Handle: TCanHandle; CanSockAddr: TCanSockAddr): cint; overload; inline;
  function  CanListen      (Handle: TCanHandle; backlog: cint): cint; inline;
  function  CanAccept      (Handle: TCanHandle; addrx: psockaddr; addrlen: psocklen): cint; inline;
  function  CanConnect     (Handle: TCanHandle; name: psockaddr; namelen: tsocklen): cint; inline; overload;
  function  CanConnect     (Handle: TCanHandle; CanSockAddr: TCanSockAddr; namelen: tsocklen): cint; inline; overload;
  function  CanConnect     (Handle: TCanHandle; CanSockAddr: TCanSockAddr): cint; inline; overload;
  function  CanShutdown    (Handle: TCanHandle; how: cint): cint; inline;
  function  CanGetSockName (Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint; inline;
  function  CanGetPeerName (Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint; inline;
  function  CanGetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: psocklen): cint; inline; overload;
  function  CanGetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: psocklen): cint; inline; overload;
  function  CanSetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: tsocklen): cint; inline; overload;
  function  CanSetSockOpt  (Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: tsocklen): cint; inline; overload;
  function  CanSocketPair  (d: cint; xtype: cint; protocol: cint; sv: pcint): cint; inline;

  // helpers for bunxh.inc
  function  CanIOCtl       (Handle: TCanHandle; Ndx: TIOCtlRequest; Data: Pointer): cint; inline; overload;
  function  CanIOCtl       (Handle: TCanHandle; Ndx: TIOCtlRequest; var Data): cint; inline; overload;
  function  CanClose       (Handle: TCanHandle): cInt; inline;

  // helpers for bunxovlh.inc
  function  CanRead        (Handle: TCanHandle; var buf; nbytes : TSize): TSsize; inline;
  function  CanRead        (Handle: TCanHandle; var CanFrame: TCanFrame): TSsize; inline;
  function  CanRead        (Handle: TCanHandle; var CanFDFrame: TCanFDFrame): TSsize; inline;
  function  CanWrite       (Handle: TCanHandle; const buf; nbytes : TSize): TSsize; inline; overload;
  function  CanWrite       (Handle: TCanHandle; const CanFrame: TCanFrame): TSsize; inline; overload;
  function  CanWrite       (Handle: TCanHandle; const CanFDFrame: TCanFDFrame): TSsize; inline; overload;

implementation

procedure TCanSockAddr.Clear;
begin
  Self.CanFamily                := 0;
  Self.CanIfIndex               := 0;
  Self.CanAddr.J1939.Name.Value := 0;
  Self.CanAddr.J1939.Addr       := 0;
  Self.CanAddr.J1939.PGN.Full   := 0;
end;

function CanSocket(domain: cint; xtype: cint; protocol: cint): TCanHandle;
begin
  Result := fpsocket(domain, xtype, protocol);
end;

function CanSocket(domain: TProtocolFamily; xtype: cint; protocol: cint): TCanHandle;
begin
  Result := fpsocket(Ord(domain), xtype, protocol);
end;

function CanSocket(xtype: cint; protocol: cint): TCanHandle;
begin
  Result := fpsocket(PF_CAN, xtype, protocol);
end;

function CanRecv(Handle: TCanHandle; buf: pointer; len: size_t; flags: cint): ssize_t;
begin
  Result := fprecv(Handle, buf, len, flags);
end;

function CanRecvFrom(Handle: TCanHandle; buf: pointer; len: size_t; flags: cint; from: psockaddr; fromlen : psocklen): ssize_t;
begin
  Result := fprecvfrom(Handle, buf, len, flags, from, fromlen);
end;

function CanSend(Handle: TCanHandle; msg: pointer; len: size_t; flags: cint): ssize_t;
begin
  Result := fpsend(Handle, msg, len, flags);
end;

function CanSendTo(Handle: TCanHandle; msg: pointer; len: size_t; flags: cint; tox: psockaddr; tolen: tsocklen): ssize_t;
begin
  Result := fpsendto(Handle, msg, len, flags, tox, tolen);
end;

function CanBind(Handle: TCanHandle; addrx: psockaddr; addrlen: tsocklen): cint;
begin
  Result := fpbind(Handle, addrx, addrlen);
end;

function CanBind(Handle: TCanHandle; CanSockAddr: TCanSockAddr; addrlen: tsocklen): cint;
begin
  Result := fpbind(Handle, @CanSockAddr, addrlen);
end;

function CanBind(Handle: TCanHandle; CanSockAddr: TCanSockAddr): cint;
begin
  Result := fpbind(Handle, @CanSockAddr, SizeOf(CanSockAddr));
end;

function CanListen(Handle: TCanHandle; backlog: cint): cint;
begin
  Result := fplisten(Handle, backlog);
end;

function CanAccept(Handle: TCanHandle; addrx: psockaddr; addrlen: psocklen): cint;
begin
  Result := fpaccept(Handle, addrx, addrlen);
end;

function CanConnect(Handle: TCanHandle; name: psockaddr; namelen: tsocklen): cint;
begin
  Result := fpconnect(Handle, name, namelen);
end;

function CanConnect(Handle: TCanHandle; CanSockAddr: TCanSockAddr; namelen: tsocklen): cint; inline; overload;
begin
  Result := fpconnect(Handle, @CanSockAddr, namelen);
end;

function CanConnect(Handle: TCanHandle; CanSockAddr: TCanSockAddr): cint; inline; overload;
begin
  Result := fpconnect(Handle, @CanSockAddr, SizeOf(CanSockAddr));
end;

function CanShutdown(Handle: TCanHandle; how: cint): cint;
begin
  Result := fpshutdown(Handle, how);
end;

function CanGetSockName(Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint;
begin
  Result := fpgetsockname(Handle, name, namelen);
end;

function CanGetPeerName(Handle: TCanHandle; name: psockaddr; namelen: psocklen): cint;
begin
  Result := fpgetpeername(Handle, name, namelen);
end;

function CanGetSockOpt(Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: psocklen): cint;
begin
  Result := fpgetsockopt(Handle, level, optname, optval, optlen);
end;

function CanGetSockOpt(Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: psocklen): cint;
begin
  Result := fpgetsockopt(Handle, level, optname, @optval, optlen);
end;

function CanSetSockOpt(Handle: TCanHandle; level: cint; optname: cint; optval: pointer; optlen: tsocklen): cint;
begin
  Result := fpsetsockopt(Handle, level, optname, optval, optlen);
end;

function CanSetSockOpt(Handle: TCanHandle; level: cint; optname: cint; var optval; optlen: tsocklen): cint;
begin
  Result := fpsetsockopt(Handle, level, optname, @optval, optlen);
end;

function CanSocketPair(d: cint; xtype: cint; protocol: cint; sv: pcint): cint;
begin
  Result := fpsocketpair(d, xtype, protocol, sv);
end;

function CanIOCtl(Handle: TCanHandle; Ndx: TIOCtlRequest; Data: Pointer): cint;
begin
  Result := fpioctl(Handle, Ndx, Data);
end;

function CanIOCtl(Handle: TCanHandle; Ndx: TIOCtlRequest; var Data): cint;
begin
  Result := fpioctl(Handle, Ndx, @Data);
end;

function CanClose(Handle : TCanHandle): cInt;
begin
  Result := fpclose(Handle);
end;

function CanRead(Handle: TCanHandle; var buf; nbytes : TSize): TSsize;
begin
  Result := fpread(Handle, buf, nbytes);
end;

function CanRead(Handle: TCanHandle; var CanFrame: TCanFrame): TSsize;
begin
  Result := fpread(Handle, @CanFrame, SizeOf(CanFrame));
end;

function CanRead(Handle: TCanHandle; var CanFDFrame: TCanFDFrame): TSsize;
begin
  Result := fpread(Handle, @CanFDFrame, SizeOf(CanFDFrame));
end;

function CanWrite(Handle: TCanHandle; const buf; nbytes : TSize): TSsize;
begin
  Result := fpwrite(Handle, buf, nbytes);
end;

function CanWrite(Handle: TCanHandle; const CanFrame: TCanFrame): TSsize;
begin
  Result := fpwrite(Handle, @CanFrame, SizeOf(CanFrame));
end;

function CanWrite(Handle: TCanHandle; const CanFDFrame: TCanFDFrame): TSsize;
begin
  Result := fpwrite(Handle, @CanFDFrame, SizeOf(CanFDFrame));
end;

end.
