unit can.isotp;

interface

uses
  ctypes, can;

(* SPDX-License-Identifier: ((GPL-2.0-only WITH Linux-syscall-note) OR BSD-3-Clause) *)
(*
 * linux/can/isotp.h
 *
 * Definitions for isotp CAN sockets (ISO 15765-2:2016)
 *
 * Copyright (c) 2020 Volkswagen Group Electronic Research
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Volkswagen nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ('GPL') version 2, in which case the provisions of the
 * GPL apply INSTEAD OF those given above.
 *
 * The provided data structures and external interfaces from this code
 * are not restricted to be used by modules with a GPL compatible license.
 *
 *  SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING  ANY WAY  OF THE USE
 * OF  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *)

const
  SOL_CAN_ISOTP = SOL_CAN_BASE + CAN_ISOTP;

  (* for socket options affecting the socket (not the global system) *)

  CAN_ISOTP_OPTS     = 1;  (* pass struct can_isotp_options             *)

  CAN_ISOTP_RECV_FC  = 2;  (* pass struct can_isotp_fc_options          *)

  (* sockopts to force stmin timer values for protocol regression tests *)

  CAN_ISOTP_TX_STMIN = 3;  (* pass __u32 value in nano secs             *)
                           (* use this time instead of value            *)
                           (* provided in FC from the receiver          *)

  CAN_ISOTP_RX_STMIN = 4;  (* pass __u32 value in nano secs             *)
                           (* ignore received CF frames which           *)
                           (* timestamps differ less than val           *)

  CAN_ISOTP_LL_OPTS  = 5;  (* pass struct can_isotp_ll_options          *)

type
  can_isotp_options = record
    flags:          cuint32;  (* set flags for isotp behaviour.       *)
                              (* __u32 value : flags see below        *)
    frame_txtime:   cuint32;  (* frame transmission time (N_As/N_Ar)  *)
                              (* __u32 value : time in nano secs      *)
    ext_address:    cuint8;   (* set address for extended addressing  *)
                              (* __u8 value : extended address        *)
    txpad_content:  cuint8;   (* set content of padding byte (tx)     *)
                              (* __u8 value : content on tx path      *)
    rxpad_content:  cuint8;	  (* set content of padding byte (rx)     *)
                              (* __u8 value : content on rx path      *)
    rx_ext_address: cuint8;   (* set address for extended addressing  *)
                              (* __u8 value : extended address (rx)   *)
  end;

  can_isotp_fc_options = record
    bs: cuint8;      (* blocksize provided in FC frame        *)
                     (* __u8 value : blocksize. 0 = off       *)
    stmin: cuint8;   (* separation time provided in FC frame  *)
                     (* __u8 value :                          *)
                     (* 0x00 - 0x7F : 0 - 127 ms              *)
                     (* 0x80 - 0xF0 : reserved                *)
                     (* 0xF1 - 0xF9 : 100 us - 900 us         *)
                     (* 0xFA - 0xFF : reserved                *)
    wftmax: cuint8;  (* max. number of wait frame transmiss.  *)
                     (* __u8 value : 0 = omit FC N_PDU WT     *)
  end;

  can_isotp_ll_options = record
    mtu: cuint8;       (* generated & accepted CAN frame type  *)
                       (* __u8 value :                         *)
                       (* CAN_MTU   (16) -> standard CAN 2.0   *)
                       (* CANFD_MTU (72) -> CAN FD frame       *)
    tx_dl: cuint8;     (* tx link layer data length in bytes   *)
                       (* (configured maximum payload length)  *)
                       (* __u8 value : 8,12,16,20,24,32,48,64  *)
                       (* => rx path supports all LL_DL values *)
    tx_flags: cuint8;  (* set into struct canfd_frame.flags    *)
                       (* at frame creation: e.g. CANFD_BRS    *)
                       (* Obsolete when the BRS flag is fixed  *)
                       (* by the CAN netdriver configuration   *)
  end;

const

  (* flags for isotp behaviour *)

  CAN_ISOTP_LISTEN_MODE   = $001;  (* listen only (do not send FC)     *)
  CAN_ISOTP_EXTEND_ADDR   = $002;  (* enable extended addressing       *)
  CAN_ISOTP_TX_PADDING    = $004;  (* enable CAN frame padding tx path *)
  CAN_ISOTP_RX_PADDING    = $008;  (* enable CAN frame padding rx path *)
  CAN_ISOTP_CHK_PAD_LEN   = $010;  (* check received CAN frame padding *)
  CAN_ISOTP_CHK_PAD_DATA  = $020;  (* check received CAN frame padding *)
  CAN_ISOTP_HALF_DUPLEX   = $040;  (* half duplex error state handling *)
  CAN_ISOTP_FORCE_TXSTMIN = $080;  (* ignore stmin from received FC    *)
  CAN_ISOTP_FORCE_RXSTMIN = $100;  (* ignore CFs depending on rx stmin *)
  CAN_ISOTP_RX_EXT_ADDR   = $200;  (* different rx extended addressing *)
  CAN_ISOTP_WAIT_TX_DONE  = $400;  (* wait for tx completion           *)


  (* default values *)

  CAN_ISOTP_DEFAULT_FLAGS        = 0;
  CAN_ISOTP_DEFAULT_EXT_ADDRESS  = $00;
  CAN_ISOTP_DEFAULT_PAD_CONTENT  = $CC;  (* prevent bit-stuffing *)
  CAN_ISOTP_DEFAULT_FRAME_TXTIME = 0;
  CAN_ISOTP_DEFAULT_RECV_BS      = 0;
  CAN_ISOTP_DEFAULT_RECV_STMIN   = $00;
  CAN_ISOTP_DEFAULT_RECV_WFTMAX  = 0;

  CAN_ISOTP_DEFAULT_LL_MTU       = CAN_MTU;
  CAN_ISOTP_DEFAULT_LL_TX_DL     = CAN_MAX_DLEN;
  CAN_ISOTP_DEFAULT_LL_TX_FLAGS  = 0;

  (*
   * Remark on CAN_ISOTP_DEFAULT_RECV_* values:
   *
   * We can strongly assume, that the Linux Kernel implementation of
   * CAN_ISOTP is capable to run with BS=0, STmin=0 and WFTmax=0.
   * But as we like to be able to behave as a commonly available ECU,
   * these default settings can be changed via sockopts.
   * For that reason the STmin value is intentionally _not_ checked for
   * consistency and copied directly into the flow control (FC) frame.
   *)

implementation

end.

