unit can;

interface

uses
  ctypes, sockets, termio;

const
  IF_NAMESIZE = 16;
  IFNAMSIZ    = IF_NAMESIZE;

  DISABLE_CAN_OPTION: integer = 0; // helper for socket options like fpsetsockopt(s, SOL_CAN_RAW, CAN_RAW_FD_FRAMES, @DISABLE_CAN_OPTION, sizeof(DISABLE_CAN_OPTION));
  ENABLE_CAN_OPTION:  integer = 1; // helper for socket options like fpsetsockopt(s, SOL_CAN_RAW, CAN_RAW_FD_FRAMES, @ENABLE_CAN_OPTION, sizeof(ENABLE_CAN_OPTION));

{$i can.ioctl.inc}
{$i can.if.inc}
{$i can.sockios.inc}

(* SPDX-License-Identifier: ((GPL-2.0-only WITH Linux-syscall-note) OR BSD-3-Clause) *)
(*
 * linux/can.h
 *
 * Definitions for CAN network layer (socket addr / CAN frame / CAN filter)
 *
 * Authors: Oliver Hartkopp <oliver.hartkopp@volkswagen.de>
 *          Urs Thuermann   <urs.thuermann@volkswagen.de>
 * Copyright (c) 2002-2007 Volkswagen Group Electronic Research
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of Volkswagen nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ('GPL') version 2, in which case the provisions of the
 * GPL apply INSTEAD OF those given above.
 *
 * The provided data structures and external interfaces from this code
 * are not restricted to be used by modules with a GPL compatible license.
 *
 *  SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * 'AS IS' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER  CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING  ANY WAY  OF THE USE
 * OF  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *)

(* controller area network (CAN) kernel definitions *)

const
  (* special address description flags for the CAN_ID *)
  CAN_EFF_FLAG: cuint32 = $80000000;	(* EFF/SFF is set in the MSB *)
  CAN_RTR_FLAG: cuint32 = $40000000;	(* remote transmission request *)
  CAN_ERR_FLAG: cuint32 = $20000000;	(* error message frame *)

  (* valid bits in CAN ID for frame formats *)
  CAN_SFF_MASK: cuint32 = $000007FF;	(* standard frame format (SFF) *)
  CAN_EFF_MASK: cuint32 = $1FFFFFFF;	(* extended frame format (EFF) *)
  CAN_ERR_MASK: cuint32 = $1FFFFFFF;	(* omit EFF, RTR, ERR flags *)

(*
 * Controller Area Network Identifier structure
 *
 * bit 0-28	: CAN identifier (11/29 bit)
 * bit 29	: error message frame flag (0 = data frame, 1 = error message)
 * bit 30	: remote transmission request flag (1 = rtr frame)
 * bit 31	: frame format flag (0 = standard 11 bit, 1 = extended 29 bit)
 *)
type
	canid_t = cuint32;

const
  CAN_SFF_ID_BITS = 11;
  CAN_EFF_ID_BITS = 29;

(*
 * Controller Area Network Error Message Frame Mask structure
 *
 * bit 0-28	: error class mask (see include/uapi/linux/can/error.h)
 * bit 29-31	: set to zero
 *)
type
	can_err_mask_t = cuint32;

const
  (* CAN payload length and DLC definitions according to ISO 11898-1 *)
  CAN_MAX_DLC  = 8;
  CAN_MAX_DLEN = 8;

  (* CAN FD payload length and DLC definitions according to ISO 11898-7 *)
  CANFD_MAX_DLC  = 15;
  CANFD_MAX_DLEN = 64;

(**
 * struct can_frame - basic CAN frame structure
 * @can_id:  CAN ID of the frame and CAN_*_FLAG flags, see canid_t definition
 * @can_dlc: frame payload length in byte (0 .. 8) aka data length code
 *           N.B. the DLC field from ISO 11898-1 Chapter 8.4.2.3 has a 1:1
 *           mapping of the 'data length code' to the real payload length
 * @__pad:   padding
 * @__res0:  reserved / padding
 * @__res1:  reserved / padding
 * @data:    CAN frame payload (up to 8 byte)
 *)
{$PACKRECORDS 8}
type 
  can_frame = record
    can_id : canid_t;  (* 32 bit CAN_ID + EFF/RTR/ERR flags *)
    can_dlc: cuint8;   (* frame payload length in byte (0 .. CAN_MAX_DLEN) *)
    __pad:   cuint8;   (* padding *)
    __res0:  cuint8;   (* reserved / padding *)
    __res1:  cuint8;   (* reserved / padding *)
    case byte of
      1: ( data:      array[0..CAN_MAX_DLEN-1]         of cuint8);  // originally defined in can.h
      2: ( DataText:  array[0..CAN_MAX_DLEN-1]         of char);    // added helper for string assignment
      3: ( DataByte:  array[0..CAN_MAX_DLEN-1]         of byte);    // added helper for byte assignment (same as cuint8)
      4: ( DataWord:  array[0..(CAN_MAX_DLEN div 2)-1] of word);    // added helper for word assignment
      5: ( DataDWord: array[0..(CAN_MAX_DLEN div 4)-1] of dword);   // added helper for double word assignment
      6: ( DataQWord: array[0..(CAN_MAX_DLEN div 8)-1] of qword);   // added helper for quad word assignment
  end;
{$PACKRECORDS DEFAULT}

(*
 * defined bits for canfd_frame.flags
 *
 * The use of struct canfd_frame implies the Extended Data Length (EDL) bit to
 * be set in the CAN frame bitstream on the wire. The EDL bit case turns
 * the CAN controllers bitstream processor into the CAN FD mode which creates
 * two new options within the CAN FD frame specification:
 *
 * Bit Rate Switch - to indicate a second bitrate is/was used for the payload
 * Error State Indicator - represents the error state of the transmitting node
 *
 * As the CANFD_ESI bit is internally generated by the transmitting CAN
 * controller only the CANFD_BRS bit is relevant for aReal CAN controllers when
 * building a CAN FD frame for transmission. Setting the CANFD_ESI bit can make
 * sense for virtual CAN interfaces to test applications with echoed frames.
 *)
const
  CANFD_BRS = $01;	(* bit rate switch (second bitrate for payload data) *)
  CANFD_ESI = $02;	(* error state indicator of the transmitting node *)

(**
 * struct canfd_frame - CAN flexible data rate frame structure
 * @can_id: CAN ID of the frame and CAN_*_FLAG flags, see canid_t definition
 * @len:    frame payload length in byte (0 .. CANFD_MAX_DLEN)
 * @flags:  additional flags for CAN FD
 * @__res0: reserved / padding
 * @__res1: reserved / padding
 * @data:   CAN FD frame payload (up to CANFD_MAX_DLEN byte)
 *)
{$PACKRECORDS 8}
type
  canfd_frame = record
    can_id: canid_t;  (* 32 bit CAN_ID + EFF/RTR/ERR flags *)
    len:    cuint8;   (* frame payload length in byte *)
    flags:  cuint8;   (* additional flags for CAN FD *)
    __res0: cuint8;   (* reserved / padding *)
    __res1: cuint8;   (* reserved / padding *)
    case byte of
      1: ( data:      array[0..CANFD_MAX_DLEN-1]         of cuint8);  // originally defined in can.h
      2: ( DataText:  array[0..CANFD_MAX_DLEN-1]         of char);    // added helper for string assignment
      3: ( DataByte:  array[0..CANFD_MAX_DLEN-1]         of byte);    // added helper for byte assignment (same as cuint8)
      4: ( DataWord:  array[0..(CANFD_MAX_DLEN div 2)-1] of word);    // added helper for word assignment
      5: ( DataDword: array[0..(CANFD_MAX_DLEN div 4)-1] of dword);   // added helper for double word assignment
      6: ( DataQword: array[0..(CANFD_MAX_DLEN div 8)-1] of qword);   // added helper for quad word assignment
  end;
{$PACKRECORDS DEFAULT}

const
  CAN_MTU   = SizeOf(can_frame);
  CANFD_MTU = SizeOf(canfd_frame);

  (* particular protocols of the protocol family PF_CAN *)
  CAN_RAW    = 1;	(* RAW sockets *)
  CAN_BCM    = 2;	(* Broadcast Manager *)
  CAN_TP16   = 3;	(* VAG Transport Protocol v1.6 *)
  CAN_TP20   = 4;	(* VAG Transport Protocol v2.0 *)
  CAN_MCNET  = 5;	(* Bosch MCNet *)
  CAN_ISOTP  = 6;	(* ISO 15765-2 Transport Protocol *)
  CAN_J1939  = 7;	(* SAE J1939 *)
  CAN_NPROTO = 8;

  SOL_CAN_BASE = 100;

type
  __kernel_sa_family_t = cushort; // word

(**
 * struct sockaddr_can - the sockaddr structure for CAN sockets
 * @can_family:  address family number AF_CAN.
 * @can_ifindex: CAN network interface index.
 * @can_addr:    protocol specific address information
 *)
  sockaddr_can = record
    can_family: __kernel_sa_family_t;
    can_ifindex: cint; // Original type was integer. Is cint (which equals cint32) a proper translation?
    can_addr: record
      case byte of
        1: (tp: record
              rx_id: canid_t;
              tx_id: canid_t;
            end;);
            
        2: (j1939: record
              (* 8 byte name when using dynamic addressing *)
              name: cuint64;

              (* pgn:
               * 8 bit: PS in PDU2 , else 0
               * 8 bit: PF
               * 1 bit: DP
               * 1 bit: reserved
               *)
              pgn: cuint32;

              (* 1 byte address *)
              addr: cuint8;
            end;);
        (* reserved for future CAN protocols address information *)
    end;
  end;

(**
 * struct can_filter - CAN ID based filter in can_register().
 * @can_id:   relevant bits of CAN ID which are not masked out.
 * @can_mask: CAN mask (see description)
 *
 * Description:
 * A filter matches, when
 *
 *          <received_can_id> and mask = can_id and mask
 *
 * The filter can be inverted (CAN_INV_FILTER bit set in can_id) or it can
 * filter for error message frames (CAN_ERR_FLAG bit set in mask).
 *)
  can_filter = record
    can_id:   canid_t;
    can_mask: canid_t;
  end;

const 
  CAN_INV_FILTER: cuint32 = $20000000;	(* to be set in can_filter.can_id *)
  CAN_RAW_FILTER_MAX = 512;	(* maximum number of can_filter set via setsockopt() *)

// helper methods:

procedure memset(AAddress: Pointer; const ANewValue: byte; const ASize: SizeInt); inline;

implementation

procedure memset(AAddress: Pointer; const ANewValue: byte; const ASize: SizeInt);
begin
  FillByte(AAddress, ASize, ANewValue);
end;

end.
